 /* 
apb_dma.c programmed by Ivan Wang
APB dma program 2004/11/11 04:24pm


APB dma function usage:
        // First, you must allocate structure and call init
            apb_dma_data_t  *priv;
            priv=apb_dma_alloc();
            //prepare variables for initialize
            priv->base=CPE_APBDMA_VA_BASE;          //base addr            
            priv->channel=PMU_SSP_DMA_CHANNEL;      //DMA channel number
            apb_dma_init(priv);
            
        // Second, add dma entry
            //prepare parameter for add dma entry
            apb_dma_parm_t parm;
            parm.src=SRC_ADDR;                      // given source phy addr
            parm.dest=DEST_ADDR;                    // given dest phy addr
            parm.width=APBDMA_WIDTH_32BIT;          // width 32 bit
            parm.req_num=REQ_SSP_NUM;               // hardware request number
            parm.sctl=APBDMA_CTL_INC2;              // source increase 2 bytes
            parm.dctl=APBDMA_CTL_FIX;               // dest fix increase
            parm.stype=APBDMA_TYPE_APB;             // indicate source is APB device
            parm.dtype=APBDMA_TYPE_AHB;             // indicate source is AHB device
            parm.size=200;                          // transfer 200 count
                                                    // if parm.sinc=2,transer size is 200*2
            parm.burst=0;                           // yes/no burst
            parm.irq=1;                             // trigger irq
            
            apb_dma_add(priv,&parm);                // call add function

        // Third, start dma
            apb_dma_start(priv);
        
        //clear interrupt
            apbdma_clear_int(priv);
                    
        // Last, Free structure
            apb_dma_free(priv);
            
        // useful function for you to maintain
            apb_dma_reset(priv); //reset dma action and stop
*/
#include <linux/module.h>
#include <linux/mm.h>
#include <linux/slab.h>
#include <asm/types.h>
#include <asm/io.h>
#include <asm/arch/apb_dma.h>


apb_dma_data_t *apb_dma_alloc(void)
{
    apb_dma_data_t *priv;
    priv=(apb_dma_data_t *)kmalloc(sizeof(apb_dma_data_t), GFP_KERNEL);
    return priv;
}

void apb_dma_free(apb_dma_data_t *priv)
{
    apb_dma_reset(priv);
    kfree(priv);
}

/* 
You must given:
    base:       base address of ahb dma
    channel:    apb dma channel
*/
int apb_dma_init(apb_dma_data_t *priv)
{   
//    int             i,size;
    
    if(priv==0)
    {
        printk("NULL priv data!\n");
        return 0;
    }
   
    priv->channel_base=(priv->base)+APBDMA_DMA_CHAN_BASE+((priv->channel)*0x10);
    printk("priv->channel_base=0x%x\n",priv->channel_base);
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_SRC_0x0)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_DEST_0x4)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CYCLE_0x8)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CMD_0xC)=0;
    return 1;
}

#define FAIL_INC    100
int apb_dma_get_inc(int inc,unsigned int burst)
{
    switch (inc)
    {
        case 0:
            return 0;
        case 1:
            return 1;
        case 2:
            return 2;
        case 4:
            if(burst)
                return 1;
            else
                return 3;
        case 8:
            if(burst)
                return 2;
            break;
        case 16:
            if(burst)
                return 3;
            break;
        case -1:
            return 5;
        case -2:
            return 6;
        case -4:
            return 7;
        default:
            break;
    }
    return FAIL_INC;
}


/*
You must prepared:
    src: source address
    dest:dest address
    sw: source width (0/1/2=>8/16/32)
    dw: dest width (0/1/2=>8/16/32)
    hmode: hardware request/grant number
    stype: source type (0/1 => APB/AHB)    
    dtype: dest type (0/1 => APB/AHB)      
    sctl:  source incremental              
    dctl:  dest incremental                
    burst: (0/1)==>(disable/enable)        
    size: dma count
    irq: (0/1)==>(disable/enable)
 */
inline void apb_dma_add(apb_dma_data_t *priv,apb_dma_parm_t *parm)
{
    unsigned int    val;
    unsigned int    sinc,dinc;
    
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_SRC_0x0)=parm->src;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_DEST_0x4)=parm->dest;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CYCLE_0x8)=parm->size;
    
    if((sinc=apb_dma_get_inc(parm->sctl,parm->burst))==FAIL_INC)
    {
        printk("Fail to add APB dma src incremental %d with burst %d\n",parm->sctl,parm->burst);
        return;
    }
    if((dinc=apb_dma_get_inc(parm->dctl,parm->burst))==FAIL_INC)
    {
        printk("Fail to add APB dma dest incremental %d with burst %d\n",parm->dctl,parm->burst);
        return;
    }
//printk("sin=0x%x dinc=0x%x\n",sinc,dinc);
    val=((((parm->width)<<20)&0x300000)|
        (((parm->req_num)<<16)&0xf0000)|
        ((dinc<<12)&0x7000)|
        ((sinc<<8)&0x700)|
        (((parm->dtype)<<7)&0x80)|
        (((parm->stype)<<6)&0x40)|
        (((parm->burst)<<3)&0x8)|
        (((parm->irq)<<2)&0x4)|     // normal intr
        (((parm->irq)<<5)&0x20));  // error intr
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CMD_0xC)=val;
}

inline void apb_dma_start(apb_dma_data_t *priv)
{
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CMD_0xC)|=0x1;
}

void apb_dma_reset(apb_dma_data_t *priv)
{
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_SRC_0x0)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_DEST_0x4)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CYCLE_0x8)=0;
    *(unsigned int *)(priv->channel_base+APBDMA_DMA_CMD_0xC)=0;
}

inline void apbdma_clear_int(apb_dma_data_t *priv)
{
    *(volatile unsigned int *)(priv->channel_base+APBDMA_DMA_CMD_0xC)&=(~0x2);
}

// John add, to get interrupt status
// Bit 0 indicate interrupt created, 
// bit 1 indicate interrupt error
unsigned int apbdma_get_status(apb_dma_data_t *priv)
{
	u32 status=0;
    u32 temp;
	u32 base_addr;

	base_addr = priv->base + APBDMA_DMA_CHAN_BASE + (APDDMA_CHANNEL_STEP * priv->channel);
    
    temp = *(volatile u32 *)(base_addr + APBDMA_DMA_CMD_0xC);

	if ( temp & APBDMA_INT_ERR_FLAG ) 
	   status |= INT_ERROR;   //error

	if ( temp & APBDMA_INT_FLAG )
	   status |= INT_TRIGGER;   //complete

    return status;
}


inline int apbdma_get_irq(unsigned int base)
{
    unsigned int     status;
	u32 base_addr;
	u32 i;

	base_addr = base+APBDMA_DMA_CHAN_BASE + APBDMA_DMA_CMD_0xC;

	for (i=0;i<4;i++) 
	{
       status=*(volatile u32 *)(base_addr + APDDMA_CHANNEL_STEP*i );
       if (status & APBDMA_INT_FLAG)
          return i;
    }

    return -1;
}

EXPORT_SYMBOL(apb_dma_alloc);
EXPORT_SYMBOL(apb_dma_free);
EXPORT_SYMBOL(apb_dma_init);
EXPORT_SYMBOL(apb_dma_add);
EXPORT_SYMBOL(apb_dma_start);
EXPORT_SYMBOL(apb_dma_reset);
EXPORT_SYMBOL(apbdma_clear_int);
//EXPORT_SYMBOL(apbdma_get_dest_addr);
EXPORT_SYMBOL(apbdma_get_status);
EXPORT_SYMBOL(apbdma_get_irq);
